import copy
import json
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional, Text, Tuple, Union

import numpy as np
from spey import ExpectationType
from spey.base import ModelConfig
from spey.system.exceptions import InvalidInput

from . import WorkspaceInterpreter, manager


class Base(ABC):
    """Base class for pyhf data input"""

    def __call__(self, expected: ExpectationType = ExpectationType.observed) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        return None, None, None

    @abstractmethod
    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """

    @property
    @abstractmethod
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""


@dataclass
class SimpleModelData(Base):
    """
    Dataclass for simple statistical model

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[float]``): observations
        absolute_uncertainties (``List[float]``): absolute uncertainties on the background
    """

    signal_yields: List[float]
    background_yields: List[float]
    data: List[float]
    absolute_uncertainties: List[float]

    def __post_init__(self):
        assert (
            len(self.signal_yields)
            == len(self.background_yields)
            == len(self.data)
            == len(self.absolute_uncertainties)
        ), "Lenght of all input must be the same."

    def __call__(self, expected: ExpectationType = ExpectationType.observed) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        model = manager.pyhf.simplemodels.uncorrelated_background(
            self.signal_yields, self.background_yields, self.absolute_uncertainties
        )

        data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        ) + model.config.auxdata

        return None, model, data

    @property
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return any(x > 0 for x in self.signal_yields)

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        signal = np.array(self.signal_yields)
        nb = np.array(self.background_yields)
        minimum_poi = (
            -np.min(np.true_divide(nb[signal != 0.0], signal[signal != 0.0])).astype(
                np.float32
            )
            if len(signal[signal != 0.0]) > 0
            else -np.inf
        )

        _, model, _ = self()

        bounds = model.config.suggested_bounds()
        bounds[model.config.poi_index] = (
            max(minimum_poi, -10.0) if allow_negative_signal else 0.0,
            bounds[model.config.poi_index][1] if not poi_upper_bound else poi_upper_bound,
        )

        return ModelConfig(
            poi_index=model.config.poi_index,
            minimum_poi=minimum_poi,
            suggested_init=model.config.suggested_init(),
            suggested_bounds=bounds,
            parameter_names=model.config.par_names,
        )


@dataclass
class FullStatisticalModelData(Base):
    """
    Data container for the full statistical model.

    Args:
        signal_patch (``List[Dict]``): Patch data for signal model. please see
            `this link <https://pyhf.readthedocs.io/en/v0.7.0/likelihood.html>`_ for details on
            the structure of the input.
        background_only_model (``Dict`` or ``Text``): This input expects background only data
            that describes the full statistical model for the background. It also accepts ``str``
            input which indicates the full path to the background only ``JSON`` file.
    """

    signal_patch: List[Dict]
    background_only_model: Union[Dict, Text]

    def __post_init__(self):
        if isinstance(self.background_only_model, str):
            if not os.path.isfile(self.background_only_model):
                raise FileNotFoundError(f"Can not find {self.background_only_model}")
            with open(self.background_only_model, "r", encoding="uft-8") as f:
                self.background_only_model = json.load(f)

        interpreter = WorkspaceInterpreter(self.background_only_model)
        interpreter.add_patch(self.signal_patch)

        # set data as expected background events
        self.expected_background_yields = interpreter.expected_background_yields

        self.workspace = manager.pyhf.Workspace(self.background_only_model)

        self._model = None
        # Initialise config
        model = self()[1]

        min_ratio = []
        for idc, channel in enumerate(self.background_only_model.get("channels", [])):
            current_signal = []
            for sigch in self.signal_patch:
                if idc == int(sigch["path"].split("/")[2]):
                    current_signal = np.array(
                        sigch.get("value", {}).get("data", []), dtype=np.float32
                    )
                    break
            if len(current_signal) == 0:
                continue
            current_bkg = []
            for ch in channel["samples"]:
                if len(current_bkg) == 0:
                    current_bkg = np.zeros(shape=(len(ch["data"]),), dtype=np.float32)
                current_bkg += np.array(ch["data"], dtype=np.float32)
            min_ratio.append(
                np.min(
                    np.true_divide(
                        current_bkg[current_signal != 0.0],
                        current_signal[current_signal != 0.0],
                    )
                )
                if np.any(current_signal != 0.0)
                else np.inf
            )
        self._minimum_poi = (
            -np.min(min_ratio).astype(np.float32) if len(min_ratio) > 0 else -np.inf
        )

        self._config = {
            "poi_index": model.config.poi_index,
            "minimum_poi": self._minimum_poi,
            "suggested_init": model.config.suggested_init(),
            "suggested_bounds": model.config.suggested_bounds(),
            "parameter_names": model.config.par_names,
            "suggested_fixed": model.config.suggested_fixed(),
        }

    @property
    def channels(self) -> Iterator[Text]:
        """Return channel names"""
        return (ch["name"] for ch in self.background_only_model["channels"])

    @property
    def channel_properties(self) -> Iterator[Tuple[int, Text, int]]:
        """Returns an iterator for channel index, name and number of bins"""
        for idx, channel in enumerate(self.channels):
            yield idx, channel, self.workspace.channel_nbins[channel]

    def __call__(
        self,
        expected: ExpectationType = ExpectationType.observed,
        include_aux: bool = True,
    ) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            include_aux (`bool`, default `True`): Include auxiliary data.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        if self._model is None:
            self._model = self.workspace.model(
                patches=[self.signal_patch],
                modifier_settings={
                    "normsys": {"interpcode": "code4"},
                    "histosys": {"interpcode": "code4p"},
                },
            )

        if expected == ExpectationType.apriori:
            try:
                data = sum(
                    (
                        self.expected_background_yields[ch]
                        for ch in self._model.config.channels
                    ),
                    [],
                )
            except KeyError as err:
                # provide a useful error message to guide the user to the solution
                missing_channels = [
                    ch
                    for ch in self._model.config.channels
                    if ch not in self.expected_background_yields
                ]
                raise InvalidInput(
                    "Unable to construct expected data. "
                    + (len(missing_channels) > 0)
                    * (
                        "\nThis is likely due to missing channels in the signal patch. "
                        + "The missing channels are: "
                        + ", ".join(missing_channels)
                        + "\nPlease provide appropriate action for the missing channels to continue."
                    )
                ) from err
            if include_aux:
                data += self._model.config.auxdata
        else:
            data = self.workspace.data(self._model, include_auxdata=include_aux)

        return self.workspace, self._model, data

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        bounds = copy.deepcopy(self._config["suggested_bounds"])
        bounds[self._config["poi_index"]] = (
            max(self._minimum_poi, -10.0) if allow_negative_signal else 0.0,
            bounds[self._config["poi_index"]][1]
            if not poi_upper_bound
            else poi_upper_bound,
        )

        return ModelConfig(
            poi_index=self._config["poi_index"],
            minimum_poi=self._minimum_poi,
            suggested_init=self._config["suggested_init"],
            suggested_bounds=bounds,
            parameter_names=self._config["parameter_names"],
            suggested_fixed=self._config["suggested_fixed"],
        )

    @property
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        for channel in (
            ch for ch in self.signal_patch if ch.get("value", None) is not None
        ):
            if any(nsig > 0.0 for nsig in channel["value"].get("data", [])):
                return True
        return False



"""Interface to convert pyhf likelihoods to simplified likelihood framework"""
import copy
import logging
import warnings
from contextlib import contextmanager
from typing import Callable, List, Literal, Optional, Text, Union

import numpy as np
import spey
import tqdm
from scipy.stats import moment, multivariate_normal
from spey.backends.default_pdf import CorrelatedBackground, ThirdMomentExpansion
from spey.optimizer.core import fit

from . import WorkspaceInterpreter
from ._version import __version__


def __dir__():
    return []


# pylint: disable=W1203, R0903

log = logging.getLogger("Spey")


class ConversionError(Exception):
    """Conversion error class"""


def make_constraint(index: int, value: float) -> Callable[[np.ndarray], float]:
    """
    Construct constraint

    Args:
        index (``int``): index of the parameter within the parameter list
        value (``float``): fixed value

    Returns:
        ``Callable[[np.ndarray], float]``:
        constraint function
    """

    def func(vector: np.ndarray) -> float:
        return vector[index] - value

    return func


@contextmanager
def _disable_logging(highest_level: int = logging.CRITICAL):
    """
    Temporary disable logging implementation, this should move into Spey

    Args:
        highest_level (``int``, default ``logging.CRITICAL``): highest level to be set in logging
    """
    previous_level = logging.root.manager.disable
    logging.disable(highest_level)
    try:
        yield
    finally:
        logging.disable(previous_level)


class Simplify(spey.ConverterBase):
    r"""
    An interface to convert pyhf full statistical model prescription into simplified likelihood
    framework as either correlated background or third moment expansion model. For details on simplified
    likelihood framework please see
    `default plug-ins page <https://spey.readthedocs.io/en/main/plugins.html#default-plug-ins>`_.

    Details on methodology can be found `in the online documentation <https://spey-pyhf.readthedocs.io/en/main/simplify.html>`_.

    Args:
        statistical_model (:obj:`~spey.StatisticalModel`): constructed full statistical model
        fittype (``Text``, default ``"postfit"``): what type of fitting should be performed ``"postfit"``
            or ``"prefit"``.
        convert_to (``Text``, default ``"default_pdf.correlated_background"``): conversion type. Should
            be either ``"default_pdf.correlated_background"`` or ``"default_pdf.third_moment_expansion"``.
        number_of_samples (``int``, default ``1000``): number of samples to be generated in order to estimate
            contract the uncertainties into a single value.
        control_region_indices (``List[int]`` or ``List[Text]``, default ``None``): indices or names of the control and
            validation regions inside the background only dictionary. For most of the cases interface will be able
            to guess the names of these regions but in case the region names are not obvious, reconstruction may
            fail thus these indices will indicate the location of the VRs and CRs within the channel list.
        include_modifiers_in_control_model (``bool``, default ``False``): This flag enables the extraction of the signal
            modifiers to be used in the control model. The control model yields will still be zero and :math:`\mu=0`
            but the contribution of the signal modifiers to the nuisance covariance matrix will be taken into account.
            By default, modifiers are excluded from the control model.
        save_model (``Text``, default ``None``): Full path to save the model details. Model will be saved as
            compressed NumPy file (``.npz``), file name should be given as ``/PATH/TO/DIR/MODELNAME.npz``.

            **Reading the saved model:**

            One can read the saved model using NumPy's :func:`load` function

            .. code:: python3

                >>> import numpy as np
                >>> saved_model = np.load("/PATH/TO/DIR/MODELNAME.npz")
                >>> data = saved_model["data"]

            This model has several containers which includes the following keywords:

            * ``"covariance_matrix"``: includes covariance matrix per bin
            * ``"background_yields"``: includes background yields per bin
            * ``"third_moments"``: (if ``convert_to="default_pdf.third_moment_expansion"``) includes third moments
              per bin
            * ``"data"``: includes observed values per bin
            * ``"channel_order"``: includes information regarding the channel order to convert a signal patch to be used
              in the simplified framework.

    Raises:
        ``ConversionError``: If the requirements are not satisfied.
        :obj:`AssertionError`: If input statistical model does not have ``pyhf`` backend or ``pyhf``
            manager does not use ``jax`` backend.

    **Example:**

    As an example, lets use the JSON files provided for ATLAS-SUSY-2019-08 analysis which can be found in
    `HEPData <https://www.hepdata.net/record/resource/1934827?landing_page=true>`_. Once these are downloaded
    one can read them as and construct a model as follows;

    .. code:: python3

        >>> import json, spey
        >>> with open("1Lbb-likelihoods-hepdata/BkgOnly.json", "r") as f:
        >>>	    background_only = json.load(f)
        >>> with open("1Lbb-likelihoods-hepdata/patchset.json", "r") as f:
        >>>     signal = json.load(f)["patches"][0]["patch"]

        >>> pdf_wrapper = spey.get_backend("pyhf")
        >>> full_statistical_model = pdf_wrapper(
        ...     background_only_model=background_only, signal_patch=signal
        ... )
        >>> full_statistical_model.backend.manager.backend = "jax"

    Note that ``patchset.json`` includes more than one patch set, thats why we used only one of them.
    The last line enables the usage of ``jax`` backend in ``pyhf`` interface which in turn enables one
    to compute Hessian of the statistical model which is needed for simplification procedure.

    Now we ca call ``"pyhf.simplify"`` model to map our full likelihood to simplified likelihood framework

    .. code:: python3

        >>> converter = spey.get_backend("pyhf.simplify")
        >>> simplified_model = converter(
        ...     statistical_model=full_statistical_model,
        ...     convert_to="default_pdf.correlated_background",
        ...     control_region_indices=[
        ...	        'WREM_cuts', 'STCREM_cuts', 'TRHMEM_cuts', 'TRMMEM_cuts', 'TRLMEM_cuts'
        ...	    ]
        ... )
        >>> print(simplified_model.backend_type)
        >>> # "default_pdf.correlated_background"
    """

    name: Text = "pyhf.simplify"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = ">=0.1.5,<0.2.0"
    """Spey version required for the backend"""

    def __call__(
        self,
        statistical_model: spey.StatisticalModel,
        fittype: Literal["postfit", "prefit"] = "postfit",
        convert_to: Literal[
            "default_pdf.correlated_background", "default_pdf.third_moment_expansion"
        ] = "default_pdf.correlated_background",
        number_of_samples: int = 1000,
        control_region_indices: Optional[Union[List[int], List[Text]]] = None,
        include_modifiers_in_control_model: bool = False,
        save_model: Optional[Text] = None,
    ) -> Union[CorrelatedBackground, ThirdMomentExpansion]:

        assert statistical_model.backend_type == "pyhf", (
            "This method is currently only available for `pyhf` full statistical models."
            + "For details please see spey-pyhf package."
        )
        assert statistical_model.backend.manager.backend == "jax", (
            "Please enable jax implementation for phyf interface. "
            + "If jax is installed this can be done by "
            + "``statistical_model.backend.manager.backend = 'jax'`` command."
        )

        bkgonly_model = statistical_model.backend.model.background_only_model
        signal_patch = statistical_model.backend.model.signal_patch

        expected = {
            "postfit": spey.ExpectationType.observed,
            "prefit": spey.ExpectationType.apriori,
        }[fittype]

        interpreter = WorkspaceInterpreter(bkgonly_model)
        bin_map = interpreter.bin_map

        # configure signal patch map with respect to channel names
        signal_patch_map, signal_modifiers_map = interpreter.patch_to_map(signal_patch)

        # Prepare a JSON patch to separate control and validation regions
        # These regions are generally marked as CR and VR
        if control_region_indices is None:
            control_region_indices = interpreter.guess_CRVR()

        if len(control_region_indices) == 0:
            raise ConversionError(
                "Can not construct control model. Please provide ``control_region_indices``."
            )

        for channel in interpreter.get_channels(control_region_indices):
            if channel in signal_patch_map and channel in signal_modifiers_map:
                interpreter.inject_signal(
                    channel,
                    [0.0] * bin_map[channel],
                    signal_modifiers_map[channel]
                    if include_modifiers_in_control_model
                    else None,
                )

        pdf_wrapper = spey.get_backend("pyhf")
        with _disable_logging():
            control_model = pdf_wrapper(
                background_only_model=bkgonly_model, signal_patch=interpreter.make_patch()
            )

        # Extract the nuisance parameters that maximises the likelihood at mu=0
        fit_opts = control_model.prepare_for_fit(expected=expected)
        _, fit_param = fit(
            **fit_opts,
            initial_parameters=None,
            fixed_poi_value=0.0,
        )

        # compute the hessian matrix without poi
        control_poi_index = fit_opts["model_configuration"].poi_index
        hessian_neg_logprob = np.delete(
            np.delete(
                -control_model.backend.get_hessian_logpdf_func(expected=expected)(
                    fit_param
                ),
                control_poi_index,
                axis=0,
            ),
            control_poi_index,
            axis=1,
        )

        # Construct multivariate normal distribution to capture correlations
        # between nuisance parameters
        nuisance_cov_matrix = np.linalg.inv(hessian_neg_logprob)
        nuisance_distribution = multivariate_normal(
            mean=np.delete(fit_param, control_poi_index), cov=nuisance_cov_matrix
        )

        # Retreive pyhf models and compare parameter maps
        if include_modifiers_in_control_model:
            stat_model_pyhf = statistical_model.backend.model()[1]
        else:
            # Remove the nuisance parameters from the signal patch
            # Note that even if the signal yields are zero, nuisance parameters
            # do contribute to the statistical model and some models may be highly
            # sensitive to the shape and size of the nuisance parameters.
            with _disable_logging():
                tmp_interpreter = copy.deepcopy(interpreter)
                for channel, data in signal_patch_map.items():
                    tmp_interpreter.inject_signal(channel=channel, data=data)
                tmp_model = spey.get_backend("pyhf")(
                    background_only_model=bkgonly_model,
                    signal_patch=tmp_interpreter.make_patch(),
                )
                stat_model_pyhf = tmp_model.backend.model()[1]
                del tmp_model, tmp_interpreter
        control_model_pyhf = control_model.backend.model()[1]
        is_nuisance_map_different = (
            stat_model_pyhf.config.par_map != control_model_pyhf.config.par_map
        )
        fit_opts = statistical_model.prepare_for_fit(expected=expected)
        suggested_fixed = fit_opts["model_configuration"].suggested_fixed
        log.debug(
            "Number of parameters to be fitted during the scan: "
            f"{fit_opts['model_configuration'].npar - len(fit_param)}"
        )

        samples = []
        warnings_list = []
        with tqdm.tqdm(
            total=number_of_samples,
            unit="sample",
            bar_format="{l_bar}{bar:20}{r_bar}{bar:-20b}",
        ) as pbar:
            while len(samples) < number_of_samples:
                nui_smp = nuisance_distribution.rvs(size=(1,))
                current_nui_params = nui_smp.tolist()
                current_nui_params.insert(control_poi_index, 0.0)
                new_params = None

                if is_nuisance_map_different:
                    constraints = []
                    new_params = np.array([np.nan] * stat_model_pyhf.config.npars)
                    for key, item in stat_model_pyhf.config.par_map.items():
                        if key in control_model_pyhf.config.par_map.keys():
                            current_params = current_nui_params[
                                control_model_pyhf.config.par_map[key]["slice"]
                            ]
                            current_range = range(
                                item["slice"].start,
                                item["slice"].stop,
                                1 if item["slice"].step is None else item["slice"].step,
                            )
                            for ival, val in zip(current_range, current_params):
                                new_params[ival] = val
                                if not suggested_fixed[ival]:
                                    constraints.append(
                                        {"type": "eq", "fun": make_constraint(ival, val)}
                                    )

                    if np.any(np.isnan(new_params)):
                        current_fit_opts = copy.deepcopy(fit_opts)
                        init_params = np.where(
                            np.isnan(new_params),
                            current_fit_opts["model_configuration"].suggested_init,
                            new_params,
                        )
                        if np.isnan(current_fit_opts["logpdf"](np.array(init_params))):
                            # if the initial value is NaN continue search
                            continue
                        current_fit_opts["constraints"] += constraints
                        with warnings.catch_warnings(record=True) as w:
                            _, new_params = fit(
                                **current_fit_opts,
                                initial_parameters=init_params.tolist(),
                                bounds=current_fit_opts[
                                    "model_configuration"
                                ].suggested_bounds,
                            )
                            warnings_list += w

                try:
                    current_sample = statistical_model.backend.get_sampler(
                        np.array(current_nui_params if new_params is None else new_params)
                    )(1, include_auxiliary=False)
                    samples.append(current_sample)
                    pbar.update()
                except ValueError:
                    # Some of the samples can lead to problems while sampling from a poisson distribution.
                    # e.g. poisson requires positive lambda values to sample from. If sample leads to a negative
                    # lambda value continue sampling to avoid that point.
                    log.debug("Problem with the sample generation")
                    log.debug(
                        f"Nuisance parameters: {current_nui_params if new_params is None else new_params}"
                    )
                    continue

        if len(warnings_list) > 0:
            log.warning(
                f"{len(warnings_list)} warning(s) generated during sampling."
                " This might be due to edge cases in nuisance parameter sampling."
            )

        samples = np.vstack(samples)

        covariance_matrix = np.cov(samples, rowvar=0)

        data = statistical_model.backend.model.workspace.data(
            stat_model_pyhf, include_auxdata=False
        )

        # NOTE: model spec might be modified within the pyhf workspace, thus
        # yields needs to be reordered properly before constructing the simplified likelihood
        signal_yields, missing_channels = [], []
        for channel_name in stat_model_pyhf.config.channels:
            try:
                signal_yields += signal_patch_map[channel_name]
            except KeyError:
                missing_channels.append(channel_name)
                signal_yields += [0.0] * bin_map[channel_name]
        if len(missing_channels) > 0:
            log.warning(
                "Following channels are not in the signal patch,"
                f" will be set to zero: {', '.join(missing_channels)}"
            )

        # NOTE background yields are first moments in simplified framework not the yield values
        # in the full statistical model!
        background_yields = np.mean(samples, axis=0)

        third_moments = []
        if convert_to == "default_pdf.correlated_background":
            backend = CorrelatedBackground(
                signal_yields=signal_yields,
                background_yields=background_yields,
                data=data,
                covariance_matrix=covariance_matrix,
            )
        elif convert_to == "default_pdf.third_moment_expansion":
            third_moments = moment(samples, moment=3, axis=0)

            backend = ThirdMomentExpansion(
                signal_yields=signal_yields,
                background_yields=background_yields,
                data=data,
                covariance_matrix=covariance_matrix,
                third_moment=third_moments,
            )
        else:
            raise ConversionError(
                "Currently available conversion methods are "
                + "'default_pdf.correlated_background', 'default_pdf.third_moment_expansion'"
            )

        if save_model is not None:
            if save_model.endswith(".npz"):
                np.savez_compressed(
                    save_model,
                    covariance_matrix=covariance_matrix,
                    background_yields=background_yields,
                    third_moments=third_moments,
                    data=data,
                    channel_order=stat_model_pyhf.config.channels,
                )

        return backend



from .manager import PyhfManager
from .helper_functions import WorkspaceInterpreter
from ._version import __version__

__all__ = ["manager", "__version__", "WorkspaceInterpreter"]


def __dir__():
    return ["__version__", "WorkspaceInterpreter"]


manager = PyhfManager()



"""Helper function for creating and interpreting pyhf inputs"""
import logging
from typing import Dict, Iterator, List, Optional, Text, Tuple, Union

__all__ = ["WorkspaceInterpreter"]


def __dir__():
    return __all__


# pylint: disable=W1203, W1201, C0103

log = logging.getLogger("Spey")


def remove_from_json(idx: int) -> Dict[Text, Text]:
    """
    Remove channel from the json file

    Args:
        idx (``int``): index of the channel

    Returns:
        ``Dict[Text, Text]``:
        JSON patch
    """
    return {"op": "remove", "path": f"/channels/{idx}"}


def add_to_json(idx: int, yields: List[float], modifiers: List[Dict]) -> Dict:
    """
    Keep channel in the json file

    Args:
        idx (``int``): index of the channel
        yields (``List[float]``): data
        modifiers (``List[Dict]``): signal modifiers

    Returns:
        ``Dict``:
        json patch
    """
    return {
        "op": "add",
        "path": f"/channels/{idx}/samples/0",
        "value": {"name": "Signal", "data": yields, "modifiers": modifiers},
    }


def _default_modifiers(poi_name: Text) -> List[Dict]:
    """Retreive default modifiers"""
    return [
        {"data": None, "name": "lumi", "type": "lumi"},
        {"data": None, "name": poi_name, "type": "normfactor"},
    ]


class WorkspaceInterpreter:
    """
    A pyhf workspace interpreter to handle book keeping for the background only models
    and convert signal yields into JSONPatch compatible for pyhf.

    Args:
        background_only_model (``Dict``): descrioption for the background only statistical model
    """

    __slots__ = [
        "background_only_model",
        "_signal_dict",
        "_signal_modifiers",
        "_to_remove",
    ]

    def __init__(self, background_only_model: Dict):
        self.background_only_model = background_only_model
        """Background only statistical model description"""
        self._signal_dict = {}
        self._signal_modifiers = {}
        self._to_remove = []

    def __getitem__(self, item):
        return self.background_only_model[item]

    @property
    def channels(self) -> Iterator[List[Text]]:
        """Retreive channel names as iterator"""
        return (ch["name"] for ch in self["channels"])

    @property
    def poi_name(self) -> Dict[Text, Text]:
        """Retreive poi name per measurement"""
        return [(mes["name"], mes["config"]["poi"]) for mes in self["measurements"]]

    @property
    def bin_map(self) -> Dict[Text, int]:
        """Get number of bins per channel"""
        return {ch["name"]: len(ch["samples"][0]["data"]) for ch in self["channels"]}

    @property
    def expected_background_yields(self) -> Dict[Text, List[float]]:
        """Retreive expected background yields with respect to signal injection"""
        yields = {}
        undefined_channels = []
        for channel in self["channels"]:
            if channel["name"] not in self.remove_list:
                yields[channel["name"]] = []
                for smp in channel["samples"]:
                    if len(yields[channel["name"]]) == 0:
                        yields[channel["name"]] = [0.0] * len(smp["data"])
                    yields[channel["name"]] = [
                        ch + dt for ch, dt in zip(yields[channel["name"]], smp["data"])
                    ]
                if channel["name"] not in self._signal_dict:
                    undefined_channels.append(channel["name"])
        if len(undefined_channels) > 0:
            log.warning(
                "Some of the channels are not defined in the patch set, "
                "these channels will be kept in the statistical model. "
            )
            log.warning(
                "If these channels are meant to be removed, please indicate them in the patch set."
            )
            log.warning(
                "Please check the following channel(s): " + ", ".join(undefined_channels)
            )
        return yields

    def guess_channel_type(self, channel_name: Text) -> Text:
        """Guess the type of the channel as CR VR or SR"""
        if channel_name not in self.channels:
            raise ValueError(f"Unknown channel: {channel_name}")
        for tp in ["CR", "VR", "SR"]:
            if tp in channel_name.upper():
                return tp

        return "__unknown__"

    def guess_CRVR(self) -> List[Text]:
        """Retreive control and validation channel names by guess"""
        return [
            name
            for name in self.channels
            if self.guess_channel_type(name) in ["CR", "VR"]
        ]

    def get_channels(self, channel_index: Union[List[int], List[Text]]) -> List[Text]:
        """
        Retreive channel names with respect to their index

        Args:
            channel_index (``List[int]``): Indices of the channels

        Returns:
            ``List[Text]``:
            Names of the channels corresponding to the given indices
        """
        return [
            name
            for idx, name in enumerate(self.channels)
            if idx in channel_index or name in channel_index
        ]

    def inject_signal(
        self, channel: Text, data: List[float], modifiers: Optional[List[Dict]] = None
    ) -> None:
        """
        Inject signal to the model

        Args:
            channel (``Text``): channel name
            data (``List[float]``): signal yields
            modifiers (``List[Dict]``): uncertainties. If None, default modifiers will be added.

        Raises:
            ``ValueError``: If channel does not exist or number of yields does not match
                with the bin size of the channel
        """
        if channel not in self.channels:
            raise ValueError(
                f"{channel} does not exist. Available channels are "
                + ", ".join(self.channels)
            )
        if len(data) != self.bin_map[channel]:
            raise ValueError(
                f"Number of bins in injection does not match to the channel. "
                f"{self.bin_map[channel]} expected, {len(data)} received."
            )

        default_modifiers = _default_modifiers(self.poi_name[0][1])
        if modifiers is not None:
            for mod in default_modifiers:
                if mod not in modifiers:
                    log.warning(
                        f"Modifier `{mod['name']}` with type `{mod['type']}` is missing"
                        f" from the input. Adding `{mod['name']}`"
                    )
                    log.debug(f"Adding modifier: {mod}")
                    modifiers.append(mod)

        self._signal_dict[channel] = data
        self._signal_modifiers[channel] = (
            default_modifiers if modifiers is None else modifiers
        )

    @property
    def signal_per_channel(self) -> Dict[Text, List[float]]:
        """Return signal yields in each channel"""
        return self._signal_dict

    def make_patch(self) -> List[Dict]:
        """
        Make a JSONPatch for the background only model

        Args:
            measurement_index (``int``, default ``0``): in case of multiple measurements
                which one to be used. Detauls is always the first measurement

        Raises:
            ``ValueError``: if there is no signal.

        Returns:
            ``List[Dict]``:
            JSONPatch file for the background only model.
        """
        if not self._signal_dict:
            raise ValueError("Please add signal yields.")

        patch = []
        to_remove = []
        for ich, channel in enumerate(self.channels):
            if channel in self._signal_dict:
                patch.append(
                    add_to_json(
                        ich, self._signal_dict[channel], self._signal_modifiers[channel]
                    )
                )
            elif channel in self._to_remove:
                to_remove.append(remove_from_json(ich))
            else:
                log.warning(f"Undefined channel in the patch set: {channel}")

        to_remove.sort(key=lambda p: p["path"].split("/")[-1], reverse=True)

        return patch + to_remove

    def reset_signal(self) -> None:
        """Clear the signal map"""
        self._signal_dict = {}
        self._to_remove = []

    def add_patch(self, signal_patch: List[Dict]) -> None:
        """Inject signal patch"""
        self._signal_dict, self._signal_modifiers, self._to_remove = self.patch_to_map(
            signal_patch=signal_patch, return_remove_list=True
        )

    def remove_channel(self, channel_name: Text) -> None:
        """
        Remove channel from the likelihood

        .. versionadded:: 0.1.5

        Args:
            channel_name (``Text``): name of the channel to be removed
        """
        if channel_name in self.channels:
            if channel_name not in self._to_remove:
                self._to_remove.append(channel_name)
        else:
            log.error(
                f"Channel {channel_name} does not exist in the background only model. "
                + "The available channels are "
                + ", ".join(list(self.channels))
            )

    @property
    def remove_list(self) -> List[Text]:
        """
        Channels to be removed from the model

        .. versionadded:: 0.1.5

        """
        return self._to_remove

    def patch_to_map(
        self, signal_patch: List[Dict], return_remove_list: bool = False
    ) -> Union[
        Tuple[Dict[Text, Dict], Dict[Text, Dict], List[Text]],
        Tuple[Dict[Text, Dict], Dict[Text, Dict]],
    ]:
        """
        Convert JSONPatch into signal map

        .. code:: python3

            >>> signal_map = {channel_name: {"data" : signal_yields, "modifiers": signal_modifiers}}


        Args:
            signal_patch (``List[Dict]``): JSONPatch for the signal
            return_remove_list (``bool``, default ``False``): Inclure channels to be removed in the output

                .. versionadded:: 0.1.5

        Returns:
            ``Tuple[Dict[Text, Dict], Dict[Text, Dict], List[Text]]`` or ``Tuple[Dict[Text, Dict], Dict[Text, Dict]]``:
            signal map including the data and modifiers and the list of channels to be removed.
        """
        signal_map, modifier_map, to_remove = {}, {}, []
        for item in signal_patch:
            path = int(item["path"].split("/")[2])
            channel_name = self["channels"][path]["name"]
            if item["op"] == "add":
                signal_map[channel_name] = item["value"]["data"]
                modifier_map[channel_name] = item["value"].get(
                    "modifiers", _default_modifiers(poi_name=self.poi_name[0][1])
                )
            elif item["op"] == "remove":
                to_remove.append(channel_name)
        if return_remove_list:
            return signal_map, modifier_map, to_remove
        return signal_map, modifier_map
